import math


def amplitude_to_db(amplitude: float, reference: float = 1e-6) -> float:
    """
    Convert amplitude from volts to decibel (dB).

    Args:
        amplitude: Amplitude in volts
        reference: Reference amplitude. Defaults to 1 µV for dB(AE)

    Returns:
        Amplitude in dB(ref)
    """
    return 20 * math.log10(amplitude / reference)


def db_to_amplitude(amplitude_db: float, reference: float = 1e-6) -> float:
    """
    Convert amplitude from decibel (dB) to volts.

    Args:
        amplitude_db: Amplitude in dB
        reference: Reference amplitude. Defaults to 1 µV for dB(AE)

    Returns:
        Amplitude in volts
    """
    return reference * 10 ** (amplitude_db / 20)



import sqlite3
from pathlib import Path
from typing import Sequence, Union

import pandas as pd

from ._database import Database, require_write_access
from ._sql import (
    QueryIterable,
    create_new_database,
    insert_from_dict,
    query_conditions,
    update_from_dict,
)
from .datatypes import FeatureRecord
from .types import SizedIterable


class TrfDatabase(Database):
    """IO Wrapper for trfdb (transient feature) database file."""

    def __init__(self, filename: str, mode: str = "ro"):
        """
        Open trfdb database file.

        Args:
            filename: Path to trfdb database file
            mode: Define database access:
                **"ro"** (read-only),
                **"rw"** (read-write),
                **"rwc"** (read-write and create empty database if it does not exist)
        """
        super().__init__(
            filename, mode=mode, table_prefix="trf", required_file_ext=".trfdb",
        )

    @staticmethod
    def create(filename: str):
        """
        Create empty trfdb.

        Args:
            filename: Path to new trfdb database file
        """
        file_schema = Path(__file__).resolve().parent / "schema_templates/trfdb.sql"
        with open(file_schema, "r", encoding="utf-8") as file:
            schema_trfdb = file.read()
        create_new_database(filename, schema_trfdb)

    def read(
        self, *, trai: Union[None, int, Sequence[int]] = None
    ) -> pd.DataFrame:
        """
        Read features to Pandas DataFrame.

        Args:
            trai: Read data by TRAI (transient recorder index)

        Returns:
            Pandas DataFrame with features
        """
        query = "SELECT * FROM trf_data " + query_conditions(isin={"TRAI": trai})
        df = pd.read_sql(query, self.connection(), index_col="TRAI")
        df.index.name = "trai"  # TRAI -> trai, as with pridb and tradb
        return df

    def iread(
        self, *, trai: Union[None, int, Sequence[int]] = None
    ) -> SizedIterable[FeatureRecord]:
        """
        Stream features with returned iterable.

        Args:
            trai: Read data by TRAI (transient recorder index)

        Returns:
            Sized iterable to sequential read features
        """
        query = "SELECT * FROM trf_data " + query_conditions(isin={"TRAI": trai})
        return QueryIterable(
            self._connection_wrapper.get_readonly_connection(),
            query,
            FeatureRecord.from_sql,
        )

    @require_write_access
    def write(self, feature_set: FeatureRecord) -> int:
        """
        Write feature record to trfdb.

        Args:
            feature_set: Feature set

        Returns:
            Index (trai) of inserted row
        """
        def convert(value):
            try:
                return float(value)
            except (ValueError, TypeError):
                return None

        with self.connection() as con:  # commit/rollback transaction
            row_dict = {
                key: convert(value)
                for key, value in feature_set.features.items()
            }
            row_dict["TRAI"] = feature_set.trai
            try:
                try:
                    return insert_from_dict(con, self._table_main, row_dict)
                except sqlite3.IntegrityError:  # UNIQUE constraint, TRAI already exists
                    # update instead
                    return update_from_dict(con, self._table_main, row_dict, "TRAI")
            except sqlite3.OperationalError:  # missing column(s)
                self._add_columns(self._table_main, list(row_dict.keys()), "REAL")
                return self.write(feature_set)  # try again



from functools import lru_cache
from pathlib import Path
from typing import Optional, Sequence, Set, Tuple, Union

import numpy as np
import pandas as pd
from tqdm import tqdm

from ._database import Database, require_write_access
from ._dataframe import iter_to_dataframe
from ._sql import (
    QueryIterable,
    create_new_database,
    insert_from_dict,
    query_conditions,
    sql_binary_search,
)
from .compression import encode_data_blob
from .datatypes import TraRecord
from .types import SizedIterable


@lru_cache(maxsize=32, typed=True)
def _create_time_vector(
    samples: int, samplerate: int, pretrigger: int = 0
) -> np.ndarray:
    t_raw = np.arange(-pretrigger, samples - pretrigger, dtype=np.int32)
    t_scaled = (t_raw / np.float32(samplerate)).astype(np.float32)
    return t_scaled


class TraDatabase(Database):
    """IO Wrapper for tradb database file."""

    def __init__(
        self, filename: str, mode: str = "ro", *, compression: bool = False,
    ):
        """
        Open tradb database file.

        Args:
            filename: Path to tradb database file
            mode: Define database access:
                **"ro"** (read-only),
                **"rw"** (read-write),
                **"rwc"** (read-write and create empty database if it does not exist)
            compression: Enable/disable FLAC compression data BLOBs for writing
        """
        super().__init__(
            filename, mode=mode, table_prefix="tr", required_file_ext=".tradb",
        )
        self._data_format = 2 if compression else 0
        self._timebase = self.globalinfo()["TimeBase"]

    @staticmethod
    def create(filename: str):
        """
        Create empty tradb.

        Args:
            filename: Path to new tradb database file
        """
        file_schema = Path(__file__).resolve().parent / "schema_templates/tradb.sql"
        with open(file_schema, "r", encoding="utf-8") as file:
            schema_tradb = file.read()
        schema_tradb = schema_tradb.format(timebase=int(1e7))  # fill placeholder / constants
        create_new_database(filename, schema_tradb)

    def channel(self) -> Set[int]:
        """Get list of channels."""
        con = self.connection()
        cur = con.execute("SELECT DISTINCT Chan FROM tr_data WHERE Chan IS NOT NULL")
        return {result[0] for result in cur.fetchall()}

    def read(self, **kwargs) -> pd.DataFrame:
        """
        Read transient data to Pandas DataFrame.

        Args:
            **kwargs: Arguments passed to `iread`

        Returns:
            Pandas DataFrame with transient data
        """
        return iter_to_dataframe(
            self.iread(**kwargs), desc="Tra", index_column="trai",
        )

    def iread(
        self,
        *,
        channel: Union[None, int, Sequence[int]] = None,
        time_start: Optional[float] = None,
        time_stop: Optional[float] = None,
        trai: Union[None, int, Sequence[int]] = None,
    ) -> SizedIterable[TraRecord]:
        """
        Stream transient data with returned Iterable.

        Args:
            channel: None if all channels should be read.
                Otherwise specify the channel number or a list of channel numbers
            time_start: Start reading at relative time (in seconds). Start at beginning if `None`
            time_stop: Stop reading at relative time (in seconds). Read until end if `None`
            trai: Read data by TRAI (transient recorder index)

        Returns:
            Sized iterable to sequential read transient data
        """
        con = self.connection()
        setid_time_start = None
        setid_time_stop = None
        setid_max = con.execute(f"SELECT MAX(SetID) FROM {self._table_main}").fetchone()[0]

        if time_start is not None:
            setid_time_start = sql_binary_search(
                connection=con,
                table="view_tr_data",
                column_value="Time",
                column_index="SetID",
                fun_compare=lambda t: t >= time_start,  # type: ignore
                lower_bound=True,  # return lower index of true conditions
            )
        if time_stop is not None:
            setid_time_stop = sql_binary_search(
                connection=con,
                table="view_tr_data",
                column_value="Time",
                column_index="SetID",
                fun_compare=lambda t: t < time_stop,  # type: ignore
                lower_bound=False,  # return upper index of true conditions
            )

        # remove upper boundary if equal to max index
        # otherwise the last row is excluded (less condition)
        if setid_time_stop == setid_max:
            setid_time_stop = None

        query = """
        SELECT vtr.*, tr.ParamID
        FROM view_tr_data vtr
        LEFT JOIN tr_data tr ON vtr.SetID == tr.SetID
        """ + query_conditions(
            isin={"vtr.Chan": channel, "vtr.TRAI": trai},
            greater_equal={"vtr.SetID": setid_time_start},
            less={"vtr.SetID": setid_time_stop},
        )
        return QueryIterable(
            self._connection_wrapper.get_readonly_connection(),
            query,
            TraRecord.from_sql,
        )

    def read_wave(
        self, trai: int, time_axis: bool = True,
    ) -> Union[Tuple[np.ndarray, np.ndarray], Tuple[np.ndarray, int]]:
        """
        Read transient signal for a given TRAI (transient recorder index).

        This method is useful in combination with `PriDatabase.read_hits`,
        that will store the TRAI in a DataFrame.

        Args:
            trai: Transient recorder index (unique key between pridb and tradb)
            time_axis: Create the correspondig time axis. Default: `True`

        Returns:
            If :attr:`time_axis` is `True`\n
            - Array with transient signal
            - Time axis

            If :attr:`time_axis` is `False`\n
            - Array with transient signal
            - Samplerate
        """
        iterable = self.iread(trai=trai)
        try:
            tra = next(iter(iterable))
        except StopIteration:
            raise ValueError("TRAI does not exists")

        if time_axis:
            return (
                tra.data,
                _create_time_vector(tra.samples, tra.samplerate, tra.pretrigger),
            )
        return tra.data, tra.samplerate

    def read_continuous_wave(
        self,
        channel: int,
        time_start: Optional[float] = None,
        time_stop: Optional[float] = None,
        *,
        time_axis: bool = True,
        show_progress: bool = True,
    ) -> Union[Tuple[np.ndarray, np.ndarray], Tuple[np.ndarray, int]]:
        """
        Read transient signal of specified channel to a single, continuous array.

        Time gaps are filled with 0's.

        Args:
            channel: Channel number to read
            time_start: Start reading at relative time (in seconds). Start at beginning if `None`
            time_stop: Stop reading at relative time (in seconds). Read until end if `None`
            time_axis: Create the correspondig time axis. Default: `True`
            show_progress: Show progress bar. Default: `True`

        Returns:
            If `time_axis` is `True`\n
            - Array with transient signal
            - Time axis

            If `time_axis` is `False`\n
            - Array with transient signal
            - Samplerate
        """
        tra_blocks = [np.empty(0, dtype=np.float32)]
        sr = 0

        iterable = self.iread(channel=channel, time_start=time_start, time_stop=time_stop)
        iterator = iter(iterable)
        if show_progress:
            iterator = tqdm(iterator, total=len(iterable), desc="Tra")

        expected_time = None
        for tra in iterator:
            # check for gaps in tra stream
            if expected_time is not None:
                time_gap = tra.time - expected_time
                if time_gap > 1e-6:
                    samples_gap = round(time_gap * tra.samplerate)
                    tra_blocks.append(
                        np.zeros(samples_gap, dtype=np.float32)
                    )

            expected_time = tra.time + float(tra.samples) / tra.samplerate

            tra_blocks.append(tra.data)
            sr = tra.samplerate

        y = np.concatenate(tra_blocks)

        if time_axis:
            return y, _create_time_vector(len(y), sr)
        return y, sr

    @require_write_access
    def write(self, tra: TraRecord) -> int:
        """
        Write transient data to pridb.

        Args:
            tra: Transient data set

        Returns:
            Index (SetID) of inserted row

        Todo:
            Status flag
        """
        # self._validate_and_update_time(tra.time)
        parameter = self._parameter(tra.param_id)
        with self.connection() as con:  # commit/rollback transaction
            return insert_from_dict(
                con,
                self._table_main,
                {
                    "Time": int(tra.time * self._timebase),
                    "Chan": int(tra.channel),
                    "Status": 32768,
                    "ParamID": int(tra.param_id),
                    "Pretrigger": int(tra.pretrigger),
                    "Thr": int(tra.threshold * 1e6 / parameter["ADC_µV"]),
                    "SampleRate": int(tra.samplerate),
                    "Samples": int(tra.samples),
                    "DataFormat": int(self._data_format),
                    "Data": encode_data_blob(tra.data, self._data_format, parameter["TR_mV"]),
                    "TRAI": (
                        int(tra.trai)
                        if tra.trai else None
                    ),
                },
            )



import collections.abc
import contextlib
import logging
import sqlite3
from functools import lru_cache
from pathlib import Path
from typing import Any, Callable, Dict, Iterator, Optional, Sequence, Tuple, TypeVar, Union

from .types import SizedIterable

logger = logging.getLogger(__name__)


class ConnectionWrapper:
    """SQLite3 connection wrapper (picklable)."""

    def __init__(self, filename: str, mode: str = "ro", multithreading: bool = False):
        # check mode
        valid_modes = ("ro", "rw", "rwc")
        if mode not in valid_modes:
            raise ValueError(f"Invalid access mode '{mode}', use: {valid_modes}")

        self._filename = str(filename)
        self._mode = mode
        self._multithreading = multithreading
        # enable multithreading for read-only connections
        if mode == "ro":
            self._multithreading = True

        self._connected = False
        self._connect()

    def _connect(self):
        """Open SQLite connection."""
        self._connection = sqlite3.connect(
            f"file:{self._filename}?mode={self._mode}",
            uri=True,
            check_same_thread=(not self._multithreading),
        )
        self._connected = True

        # set pragmas for write-mode
        if self._mode != "ro":
            self._connection.executescript(
                """
                PRAGMA journal_mode = WAL;
                PRAGMA synchronous = OFF;
                """
            )

    @property
    def filename(self) -> str:
        return self._filename

    @property
    def mode(self) -> str:
        return self._mode

    @property
    def connected(self) -> bool:
        return self._connected

    def connection(self) -> sqlite3.Connection:
        """
        Get SQLite connection object.

        Raises:
            RuntimeError: If connection is closed
        """
        if not self._connected:
            raise RuntimeError("Not connected to SQLite database")
        return self._connection

    def get_readonly_connection(self) -> "ConnectionWrapper":
        """
        Return read-only ConnectionWrapper.

        Create new connection if mode != ro.
        """
        if self._mode == "ro":
            return self
        return ConnectionWrapper(self._filename, mode="ro")

    def close(self):
        if self._connected:
            self._connection.commit()  # commit remaining changes
            self._connection.close()
            self._connected = False

    def __del__(self):
        self.close()

    def __getstate__(self):
        # commit changes, database will be reopened with __setstate__
        if self._connected:
            self._connection.commit()
        state = self.__dict__.copy()
        del state["_connection"]  # remove the unpicklable sqlite3.connection
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        # reopen connection if connected before
        if self._connected:
            self._connect()


T = TypeVar("T")
class QueryIterable(SizedIterable[T]):
    """
    Sized iterable to query results from SQLite as dictionaries.

    SQLite connection is stored in picklable ConnectionWrapper to be used with multiprocessing.
    """
    def __init__(
        self,
        connection_wrapper: ConnectionWrapper,
        query: str,
        dict_to_type: Callable[[Dict[str, Any]], T],
    ):
        self._connection_wrapper = connection_wrapper
        self._query = query
        self._dict_to_type = dict_to_type
        self._count_result: Optional[int] = None  # cache result of __len__

    def __len__(self) -> int:
        if self._count_result is None:
            self._count_result = count_sql_results(
                self._connection_wrapper.connection(),
                self._query
            )
        return self._count_result

    def __iter__(self) -> Iterator[T]:
        if self.__len__() == 0:
            logger.debug("Empty SQLite query")

        for row in read_sql_generator(self._connection_wrapper.connection(), self._query):
            yield self._dict_to_type(row)


TIsIn = Dict[str, Union[float, Sequence[float], None]]
TComparison = Dict[str, Optional[float]]


def query_conditions(
    *,
    isin: Optional[TIsIn] = None,
    equal: Optional[TComparison] = None,
    less: Optional[TComparison] = None,
    less_equal: Optional[TComparison] = None,
    greater: Optional[TComparison] = None,
    greater_equal: Optional[TComparison] = None,
) -> str:
    cond = []

    if isin is not None:
        for key, values in isin.items():
            if values is None:
                continue
            if not isinstance(values, collections.abc.Sequence):
                values = (values,)
            cond.append(
                "{:s} IN ({:s})".format(key, ", ".join(str(value) for value in values))
            )

    comparison = {
        "==": equal,
        "<": less,
        "<=": less_equal,
        ">": greater,
        ">=": greater_equal,
    }

    for comp_operator, comp_dict in comparison.items():
        if comp_dict is not None:
            for key, value in comp_dict.items():
                if value is None:
                    continue
                if isinstance(value, str):
                    value = f"'{value}'"  # add quotation marks
                cond.append(f"{key} {comp_operator} {value}")

    return "WHERE " + " AND ".join(cond) if cond else ""


def read_sql_generator(
    connection: sqlite3.Connection, query: str
) -> Iterator[Dict[str, Any]]:
    """
    Generator to query data from a SQLite connection as a dictionary.

    Args:
        connection: SQLite3 connection object
        query: SELECT Query

    Yields:
        Row of the query result set as namedtuple
    """
    cur = connection.execute(query)
    columns = [column[0] for column in cur.description]

    while True:
        values = cur.fetchone()
        if values is None:
            break
        yield dict(zip(columns, values))


def count_sql_results(connection: sqlite3.Connection, query: str) -> int:
    count_query = f"SELECT COUNT(*) FROM ({query})"
    cur = connection.execute(count_query)
    return cur.fetchone()[0]


def sql_binary_search(
    connection: sqlite3.Connection,
    table: str,
    column_value: str,
    column_index: str,
    fun_compare: Callable[[float], bool],
    *,
    lower_bound: bool = True,
) -> Optional[int]:
    """
    Helper function to find the boundary row-index for given condition on a sorted column.

    Especially using conditions on the pridb's and tradb's Time column are expensive (not indexed).
    E.g.: SELECT * FROM view_tr_data WHERE Time > 10 AND Time < 100
    Because the Time column is monotonic increasing, a fast binary search can be applied.

    Args:
        connection: SQLite connection
        table: Table name
        column_value: Name of the sorted column, e.g. Time
        column_index: Name of the indexed column
        fun_compare: Lambda function of the condition, e.g. `lambda t: t > 10` (Time > 10)
        lower_bound: Specify which index to return if condition is true for both lower/upper bound.
            Default: Return lower bound (`True`)
    """

    def get_value(index):
        cur = connection.execute(
            f"SELECT {column_value} FROM {table} WHERE {column_index} == {index}"
        )
        return cur.fetchone()[0]

    i_min, i_max = connection.execute(
        f"SELECT MIN({column_index}), MAX({column_index}) FROM {table}"
    ).fetchone()

    while True:
        v_min = get_value(i_min)
        v_max = get_value(i_max)
        if v_min > v_max:
            raise ValueError(f"Value column {column_value} not sorted")
        c_min = fun_compare(v_min)
        c_max = fun_compare(v_max)

        if c_min and c_max:  # condition true for both limits
            return i_min if lower_bound else i_max
        if not c_min and not c_max:  # condition false for both limits
            return None

        if i_max - i_min < 2:
            return i_min if c_min is True else i_max

        i_mid = (i_max + i_min) // 2
        c_mid = fun_compare(get_value(i_mid))

        if c_mid == c_min:
            i_min = i_mid
        else:
            i_max = i_mid


def create_new_database(filename: str, schema: str):
    if Path(filename).resolve().exists():
        raise ValueError("Can not create new database. File already exists")

    # open database in read-write-create mode
    with contextlib.closing(
        sqlite3.connect(f"file:{filename}?mode=rwc", uri=True)
    ) as con:
        con.executescript(schema)


def remove_none_values_from_dict(dictionary: Dict[Any, Any]):
    """Helper function to remove None values from dict."""
    return {k: v for k, v in dictionary.items() if v is not None}


@lru_cache(maxsize=128, typed=True)
def generate_insert_query(table: str, columns: Tuple[str, ...]) -> str:
    """
    Generate INSERT query with named placeholders.

    e.g.: INSERT INTO ae_data (SetID, Time, Channel) VALUES (:SetID, :Time, :Channel)

    Args:
        table: Table name
        columns: Tuple of column names
            (must be of type tuple to be hashable for caching)

    Returns:
        Query string with named placeholders
    """
    query = "INSERT INTO {table} ({columns}) VALUES ({placeholder})".format(
        table=table,
        columns=", ".join(columns),
        placeholder=", ".join([":" + col for col in columns]),
    )
    return query


def insert_from_dict(
    connection: sqlite3.Connection,
    table: str,
    row_dict: Dict[str, Any],
) -> int:
    """INSERT row for given dict of column names -> values in SQLite table."""
    row_dict = remove_none_values_from_dict(row_dict)
    columns = tuple(row_dict.keys())
    query = generate_insert_query(table, columns)
    cur = connection.execute(query, row_dict)
    return cur.lastrowid


@lru_cache(maxsize=128, typed=True)
def generate_update_query(table: str, columns: Tuple[str, ...], key_column: str) -> str:
    """
    Generate UPDATE query with named placeholders.

    e.g.: UPDATE ae_data SET Time = :Time, Channel = :Channel WHERE SetID == :SetID

    Args:
        table: Table name
        columns: Tuple of column names
            (must be of type tuple to be hashable for caching)
        key_column: Column name for WHERE clause

    Returns:
        Query string with named placeholders
    """
    columns_list = list(columns)
    try:
        columns_list.remove(key_column)
    except ValueError:
        raise ValueError(f"Argument key_column '{key_column}' must be a key of row_dict")

    query = "UPDATE {table} SET {set} WHERE {condition}".format(
        table=table,
        set=", ".join([f"{col} = :{col}" for col in columns_list]),
        condition=f"{key_column} == :{key_column}",
    )
    return query


def update_from_dict(
    connection: sqlite3.Connection,
    table: str,
    row_dict: Dict[str, Any],
    key_column: str,
) -> int:
    """UPDATE row for given key and dict of column names -> values in SQLite table."""
    row_dict = remove_none_values_from_dict(row_dict)
    columns = tuple(row_dict.keys())
    query = generate_update_query(table, columns, key_column)
    cur = connection.execute(query, row_dict)
    return cur.lastrowid



from typing import Optional, Tuple

import pandas as pd
from tqdm import tqdm

from .types import SizedIterable


def _convert_to_nullable_types(df: pd.DataFrame):
    # types not hashable -> use str repr
    dtype_mapping = {
        "int32": pd.Int32Dtype(),
        "int64": pd.Int64Dtype(),
    }
    return df.apply(
        lambda x: x.astype(
            dtype_mapping.get(
                str(x.dtype),  # try get new dtype
                x.dtype,  # return original dtype if key not found
            )
        )
    )


def iter_to_dataframe(
    iterable: SizedIterable[Tuple],
    show_progress: bool = True,
    desc: str = "",
    index_column: Optional[str] = None,
) -> pd.DataFrame:
    """
    Helper function to save iterator results in Pandas DataFrame.

    Args:
        iterable: Iterable generating `NamedTuple`s
        show_progress: Show progress bar. Default: `True`
        desc: Description shown left to the progress bar
        index_column: Set column as index. Default: `None`
    Returns:
        Pandas DataFrame
    """
    iterator = iter(iterable)
    if show_progress:
        iterator = tqdm(iterator, total=len(iterable), desc=desc)
    df = pd.DataFrame(iterator)
    if not df.empty:
        df = df.dropna(axis="columns", how="all")  # drop empty columns
        if index_column is not None:
            df = df.set_index(index_column)
    df = _convert_to_nullable_types(df)
    return df



"""
Timerpicker
===========

The determination of signal arrival times has a major influence on the localization accuracy.
Usually, arrival times are determined by the first threshold crossing (either fixed or adaptive).
Following popular methods have been proposed in the past to automatically pick time of arrivals:

.. autosummary::
    :toctree: generated

    hinkley
    aic
    energy_ratio
    modified_energy_ratio
"""

# flake8: noqa

from .timepicker import *

__all__ = [_ for _ in dir() if not _.startswith("_")]



"""
Features
========

Acoustic Emission
-----------------

.. autosummary::
    :toctree: generated

    peak_amplitude
    peak_amplitude_index
    is_above_threshold
    first_threshold_crossing
    rise_time
    energy
    signal_strength
    counts
    rms

Conversion
----------

.. autosummary::
    :toctree: generated

    amplitude_to_db
    db_to_amplitude
"""

# flake8: noqa

from .acoustic_emission import *
from .conversion import *

__all__ = [_ for _ in dir() if not _.startswith("_")]



from setuptools import setup, find_packages
from os import path

HERE = path.abspath(path.dirname(__file__))

with open(path.join(HERE, "README.md"), encoding="utf-8") as f:
    LONG_DESCRIPTION = f.read()

INSTALL_REQUIRES = [
    "numpy",
    "scipy",
    "pandas>=0.24",
    "soundfile",
    "numba",
    "tqdm",
    "typing_extensions",
]

EXTRAS_REQUIRE = {
    "docs": [
        "sphinx<2.3",
        "sphinx-autodoc-typehints==1.10.3",  # >1.10 requires sphinx>=3.0 -> to be updated soon
        "sphinx-rtd-theme",
        "sphinx-gallery",
        "pillow",  # required by sphinx-gallery
        "matplotlib",  # used in examples
    ],
    "tests": [
        "pytest",
        "coverage<5.0",
    ],
    "tools": [
        "tox",
        "pylint",
        "mypy",
    ]
}

EXTRAS_REQUIRE["dev"] = EXTRAS_REQUIRE["docs"] + EXTRAS_REQUIRE["tests"] + EXTRAS_REQUIRE["tools"]

setup(
    name="vallenae",
    version="0.2.2",
    description="Extract and analyze Acoustic Emission measurement data",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/vallen-systems/pyVallenAE",
    author="Daniel Altmann, Lukas Berbuer (Vallen Systeme GmbH)",
    author_email="software@vallen.de",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Operating System :: OS Independent",
    ],
    keywords="vallen acoustic-emission amsy sqlite pridb tradb",
    packages=find_packages("src"),
    package_dir={"": "src"},
    python_requires=">=3.6",
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS_REQUIRE,
    package_data={
        "vallenae": ["io/schema_templates/*.sql"],
    },
    project_urls={
        "Bug Reports": "https://github.com/vallen-systems/pyVallenAE/issues",
        "Source": "https://github.com/vallen-systems/pyVallenAE",
    },
)



import io

import numpy as np
from numba import njit

try:
    import soundfile as sf
    _FLAC_CODEC = True
except OSError:
    _FLAC_CODEC = False


_II16_MIN = np.iinfo(np.int16).min
_II16_MAX = np.iinfo(np.int16).max


def _check_flac_codec():
    if not _FLAC_CODEC:
        raise ValueError("FLAC library not found. Please install libsndfile.")


@njit
def _convert_to_int16(arr: np.ndarray):
    result = np.empty_like(arr, dtype=np.int16)
    for i, value in enumerate(arr):
        # clip to range of int16
        value = min(value, _II16_MAX)
        value = max(value, _II16_MIN)
        # round to next integer
        result[i] = round(value)
    return result


def decode_data_blob(
    data_blob: bytes, data_format: int, factor_millivolts: float
) -> np.ndarray:
    """
    Decodes (compressed) 16-bit ADC values from BLOB to array of voltage values.

    Args:
        data_blob: Blob from tradb
        data_format:
            - 0: uncompressed
            - 2: FLAC compression
        factor_millivolts: Factor from int16 representation to millivolts.
            Stored in tradb -> tr_params as 'TR_mV'

    Returns:
        Array of voltage values
    """
    if data_format == 0:  # uncompressed
        data = np.frombuffer(data_blob, dtype=np.int16)
        factor = 1e-3 * factor_millivolts
        return np.multiply(data, factor, dtype=np.float32)
    if data_format == 2:  # flac
        _check_flac_codec()
        data, _ = sf.read(io.BytesIO(data_blob), dtype=np.float32)
        factor_libsoundfile = 2 ** 15  # libsoundfile normalizes to +-1
        factor = 1e-3 * factor_millivolts * factor_libsoundfile
        return np.multiply(data, factor, dtype=np.float32)
    raise ValueError("Data format not supported")


def encode_data_blob(
    data: np.ndarray, data_format: int, factor_millivolts: float
) -> bytes:
    """
    Encodes array of voltage values to BLOB of 16-bit ADC values
    for memory-efficient storage.

    Args:
        data: Array with voltage values
        data_format:
            - 0: uncompressed
            - 2: FLAC compression
        factor_millivolts: Factor from int16 representation to millivolts.
            Stored in tradb -> tr_params as 'TR_mV'

    Returns:
        Data blob
    """
    data = data * 1e3 / factor_millivolts
    data_int16 = _convert_to_int16(data)
    if data_format == 0:  # uncompressed
        return data_int16.tobytes()
    if data_format == 2:  # flac
        _check_flac_codec()
        buffer = io.BytesIO()
        sf.write(buffer, data_int16, 1, format="FLAC")  # samplerate = 1
        return buffer.getvalue()
    raise ValueError("Data format not supported")
